package be.iminds.ilabt.jfed.testing.base;

import be.iminds.ilabt.jfed.call_log_output.HtmlLogOutput;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.util.JFedVersionInfo;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import org.rendersnake.HtmlCanvas;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.adapters.XmlAdapter;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;
import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.*;

import static org.rendersnake.HtmlAttributesFactory.class_;

/**
 * ApiTestResult: list of test method results
 */
public class ApiTestResult extends HtmlLogOutput {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ApiTestResult.class);

    protected final TargetAuthority testedAuthority;
    protected final GeniUser user;


    @XmlRootElement(name = "method")
    @XmlJavaTypeAdapter(JAXBMutableApiTestMethodResultAdaptor.class)
    public static class ApiTestMethodResult {
        @Nonnull final String methodName;
        @Nullable final String description;
        @Nonnull final List<String> hardDependencies;
        @Nonnull final List<String> softDependencies;

        @Nonnull final List<LogEntry> logLines;
        @Nonnull final List<SerializableApiCallDetails> apiCallDetails;
        @Nullable Throwable exception;

        @Nullable protected TestResultState state;

        /**
         * Used to add extra data to the fedmon test that run this automated test.
         * The data is added directly to the Result created by the fedmon test.
         *    (key can be nested, ex: slice -> urn -> "urn:publicID...") added to Result.results
         * */
        public static class FedmonResultExtra {
            @Nonnull private final List<String> nestedKey;
            @Nonnull private final Object value;
            private final boolean large;
            @Nonnull private final String httpMediaType; //ex: text/plain  application/json  application/xml  text/html

            public FedmonResultExtra(@Nonnull String key, @Nonnull Object value,
                                     boolean large,
                                     @Nonnull String httpMediaType) {
                this.nestedKey = Collections.singletonList(key);
                this.value = value;
                this.large = large;
                this.httpMediaType = httpMediaType;
            }
            public FedmonResultExtra(@Nonnull List<String> nestedKey, @Nonnull Object value,
                                     boolean large,
                                     @Nonnull String httpMediaType) {
                assert !nestedKey.isEmpty();
                this.nestedKey = nestedKey;
                this.value = value;
                this.large = large;
                this.httpMediaType = httpMediaType;
            }

            @Nonnull
            public List<String> getNestedKey() {
                return nestedKey;
            }
            @Nonnull
            public String getLastNestedKey() {
                assert !nestedKey.isEmpty();
                return nestedKey.get(nestedKey.size()-1);
            }

            @Nonnull
            public Object getValue() {
                return value;
            }

            @Nonnull
            public String getHttpMediaType() {
                return httpMediaType;
            }

            /**
             * @return if true, this is a large object that should be written to a file (and not added to Result by fedmon)
             */
            public boolean isLarge() {
                return large;
            }
        }
        public static class FedmonResultExtraBuilder {
            private Object value;
            private boolean large = false;
            private String httpMediaType = "text/plain";
            private List<String> nestedKey;

            public FedmonResultExtraBuilder setKey(@Nonnull String key) {
                this.nestedKey = Collections.singletonList(key);
                assert hasValidKey();
                return this;
            }
            public FedmonResultExtraBuilder setKey(@Nonnull String ... keys) {
                this.nestedKey = Arrays.asList(keys);
                assert hasValidKey();
                return this;
            }
            public FedmonResultExtraBuilder setKey(@Nonnull List<String> nestedKey) {
                this.nestedKey = nestedKey;
                assert hasValidKey();
                return this;
            }
            public boolean hasValidKey() {
                if (nestedKey == null)
                    return false;
                if (nestedKey.isEmpty())
                    return false;
                return nestedKey.stream().noneMatch(Objects::isNull);
            }

            public FedmonResultExtraBuilder setValue(@Nonnull Object value) {
                this.value = value;
                return this;
            }

            public FedmonResultExtraBuilder setLarge(boolean large) {
                this.large = large;
                return this;
            }

            /**
             * not mandatory, since it doesn't make sense for non large objects
             * @param httpMediaType
             * @return
             */
            public FedmonResultExtraBuilder setHttpMediaType(@Nonnull String httpMediaType) {
                this.httpMediaType = httpMediaType;
                return this;
            }

            public FedmonResultExtraBuilder setHttpMediaTypeToXml() {
                this.httpMediaType = "application/xml";
                return this;
            }
            public FedmonResultExtraBuilder setHttpMediaTypeToText() {
                this.httpMediaType = "text/plain";
                return this;
            }
            public FedmonResultExtraBuilder setHttpMediaTypeToJson() {
                this.httpMediaType = "application/json";
                return this;
            }
            public FedmonResultExtraBuilder setHttpMediaTypeToHtml() {
                this.httpMediaType = "text/html";
                return this;
            }

            public FedmonResultExtra create() {
                if (nestedKey == null)
                    throw new IllegalStateException("Not all required fields have been filled in (key missing)");
                if (value == null)
                    throw new IllegalStateException("Not all required fields have been filled in (value missing)");
                if (httpMediaType == null)
                    throw new IllegalStateException("Not all required fields have been filled in (httpMediaType missing)");
                if (!hasValidKey())
                    throw new IllegalStateException("Invalid nested key filled in: "+nestedKey);
                return new FedmonResultExtra(nestedKey, value, large, httpMediaType);
            }
        }

        @Nonnull protected List<FedmonResultExtra> fedmonResultExtras;

        final long startTimeMs;
        long stopTimeMs;
        long timeMs;

        public ApiTestMethodResult(
                @Nonnull String methodName,
                @Nullable String description,
                @Nonnull List<String> hardDependencies,
                @Nonnull List<String> softDependencies,
                long startTimeMs) {
            this.methodName = methodName;
            this.description = description;
            this.hardDependencies = new ArrayList<>(hardDependencies);
            this.softDependencies = new ArrayList<>(softDependencies);
            this.logLines  = new ArrayList<>();
            this.apiCallDetails  = new ArrayList<>();
            this.startTimeMs = startTimeMs;
            this.fedmonResultExtras = new ArrayList<>();
            this.exception = null;
        }

        public ApiTestMethodResult(
                @Nonnull String methodName,
                @Nullable String description,
                @Nonnull List<String> hardDependencies,
                @Nonnull List<String> softDependencies,
                @Nonnull List<LogEntry> logLines,
                @Nonnull List<SerializableApiCallDetails> apiCallDetails,
                @Nullable Throwable exception,
                @Nullable TestResultState state,
                long timeMs,
                long startTimeMs,
                long stopTimeMs) {
            this.methodName = methodName;
            this.description = description;
            this.hardDependencies = new ArrayList<>(hardDependencies);
            this.softDependencies = new ArrayList<>(softDependencies);
            this.logLines = logLines;
            this.apiCallDetails = apiCallDetails;
            this.exception = exception;
            this.state = state;
            this.timeMs = timeMs;
            this.startTimeMs = startTimeMs;
            this.stopTimeMs = stopTimeMs;
            this.fedmonResultExtras = new ArrayList<>();
        }

        public void addLogLine(@Nonnull LogLineType type, @Nullable String text, @Nullable Throwable t, boolean formatAsCode) {
            assert type != null;
            if (text == null)
                text = "";
            assert text != null;
            logLines.add(new LogEntry(type, text, t, formatAsCode));
        }
        public void addLogLine(@Nonnull LogLineType type, @Nullable String text, @Nullable Throwable t) {
            assert type != null;
            if (text == null)
                text = "";
            assert text != null;
            logLines.add(new LogEntry(type, text, t));
        }
        public void addLogLine(@Nonnull LogLineType type, @Nullable String text, boolean formatAsCode) {
            assert type != null;
            if (text == null)
                text = "";
            assert text != null;
            logLines.add(new LogEntry(type, text, formatAsCode));
        }
        public void addLogLine(@Nonnull LogLineType type, @Nullable String text) {
            assert type != null;
            if (text == null)
                text = "";
            assert text != null;
            logLines.add(new LogEntry(type, text));
        }
        public void addLogLine(@Nonnull LogEntry logEntry) {
            logLines.add(logEntry);
        }

        @XmlElement(name = "state")
        @Nullable public TestResultState getState() {
            return state;
        }

        @XmlElement(name = "methodName")
        @Nonnull public String getMethodName() {
            return methodName;
        }

        @XmlElement(name = "description")
        @Nullable public String getDescription() {
            return description;
        }

        @Nonnull public List<String> getHardDependencies() {
            return Collections.unmodifiableList(hardDependencies);
        }

        @Nonnull public List<String> getSoftDependencies() {
            return Collections.unmodifiableList(softDependencies);
        }

        @XmlElementWrapper(name = "loglines")
        @XmlElement(name = "logline")
        @Nonnull public List<LogEntry> getLogLines() {
            return Collections.unmodifiableList(logLines);
        }

        @XmlElementWrapper(name = "apicalls")
        @XmlElement(name = "apicall")
        @Nonnull public List<SerializableApiCallDetails> getApiCallDetails() {
            return Collections.unmodifiableList(apiCallDetails);
        }

        @Nullable public Throwable getException() {
            return exception;
        }

        @XmlElement(name = "durationInMs")
        public long getTimeMs() {
            return timeMs;
        }

        @XmlElement(name = "startTimeInMs")
        public long getStartTimeMs() {
            return startTimeMs;
        }

        @XmlElement(name = "stopTimeInMs")
        public long getStopTimeMs() {
            return stopTimeMs;
        }

//        public void addExtraResult(@Nonnull Object value, @Nonnull String key, @Nonnull String httpMediaType) {
//            fedmonResultExtras.add(new FedmonResultExtraBuilder().setKey(key).setValue(value).setLarge(false).setHttpMediaType(httpMediaType).createFedmonResultExtra());
//        }
//        public void addExtraResult(@Nonnull Object value, @Nonnull String ... keys, @Nonnull String httpMediaType) {
//            assert keys.length > 0;
//            fedmonResultExtras.add(new FedmonResultExtraBuilder().setNestedKey(Arrays.asList(keys)).setValue(value).setLarge(false).setHttpMediaType(httpMediaType).createFedmonResultExtra());
//        }
//        public void addLargeExtraResult(@Nonnull Object value, @Nonnull String ... keys) {
//            assert keys.length > 0;
//            fedmonResultExtras.add(new FedmonResultExtra(Arrays.asList(keys), value, true));
//        }
//        public void addExtraResult(@Nonnull List<String> nestedKey, @Nonnull Object value) {
//            assert !nestedKey.isEmpty();
//            fedmonResultExtras.add(new FedmonResultExtra(nestedKey, value, false));
//        }
        public void addExtraResult(FedmonResultExtra extra) {
            fedmonResultExtras.add(extra);
        }
        public void addExtraResult(FedmonResultExtraBuilder extra) {
            fedmonResultExtras.add(extra.create());
        }

        @Nonnull
        public List<FedmonResultExtra> getFedmonResultExtras() {
            return Collections.unmodifiableList(fedmonResultExtras);
        }

        @Override
        public String toString() {
            return "ApiTestMethodResult{" +
                    "methodName='" + methodName + '\'' +
                    ", description='" + description + '\'' +
                    ", hardDependencies=" + hardDependencies +
                    ", softDependencies=" + softDependencies +
                    ", logLines=" + logLines +
                    ", apiCallDetails=" + apiCallDetails +
                    ", exception=" + exception +
                    ", state=" + state +
                    ", timeMs=" + timeMs +
                    ", startTimeMs=" + startTimeMs +
                    ", stopTimeMs=" + stopTimeMs +
                    '}';
        }
    }


    private final Class<? extends ApiTest> testClass;
    private final String testGroup;

    public ApiTestResult(TargetAuthority testedAuthority, GeniUser user, Class<? extends ApiTest> testClass, String testGroup) {
        super("Test Result");
        this.testedAuthority = testedAuthority;
        this.user = user;

        this.testClass = testClass;
        this.testGroup = testGroup;
    }

    public Class<? extends ApiTest> getTestClass() {
        return testClass;
    }

    /**
     * ordered list of results
     */
    final List<ApiTestMethodResult> resultList = new ArrayList<>();
    int returnValue = 0;

    public List<ApiTestMethodResult> getResultList() {
        return Collections.unmodifiableList(resultList);
    }

    public Date getStartDate() {
        Date start = null;
        for (ApiTestMethodResult result : resultList) {
            Date date = new Date(result.getStartTimeMs());
            if (start == null || start.after(date))
                start = date;
        }
        return start;
    }

    public Date getStopDate() {
        Date stop = null;
        for (ApiTestMethodResult result : resultList) {
            Date date = new Date(result.getStopTimeMs());
            if (stop == null || stop.before(date))
                stop = date;
        }
        return stop;
    }

    public long getDurationMs() {
        Date begin = getStartDate();
        Date end = getStopDate();
        if (begin == null || end == null)
            return 0;
        return end.getTime() - begin.getTime();
    }

    public String getVersionString() {
        StringBuilder version = new StringBuilder();
        try {
            final String VERSION_PROPERTIES_FILE = "/version.properties";
            final String PROP_VERSION = "version";
            final String PROP_GIT_COMMIT = "build.git_commit";
            final String PROP_GIT_BRANCH = "build.git_branch";
            final String PROP_BUILD_JENKINS = "build.jenkins_build";

            Properties prop = new Properties();
            //now load the version-properties file
            prop.load(getClass().getResourceAsStream(VERSION_PROPERTIES_FILE));

            //now load the info into the disabled textfield

            version.append(prop.getProperty(PROP_VERSION, "Unknown"));
            final String jenkinsBuildNr = prop.getProperty(PROP_BUILD_JENKINS);
            if (jenkinsBuildNr != null && !jenkinsBuildNr.isEmpty())
                version.append("  - build #").append(jenkinsBuildNr);

            final String gitCommit = prop.getProperty(PROP_GIT_COMMIT);
            if (gitCommit != null && !gitCommit.isEmpty())
                version.append("  - git commit #").append(gitCommit);

            final String gitBranch = prop.getProperty(PROP_GIT_BRANCH);
            if (gitBranch != null && !gitBranch.isEmpty())
                version.append(" on ").append(gitBranch);
            System.out.println("Version: "+version.toString());
        } catch (Exception e) {
            //ignore
            version.append(" === EXCEPTION finding version: ").append(e.getMessage()).append("\"");
            LOG.warn("EXCEPTION finding version: \""+e.getMessage()+"\"", e);
        }
        return version.toString();
    }

    public String getEnvironment() {
        return JFedVersionInfo.getEnvironmentString();
    }
    /*
     * return value meaning:
     * 2: errors
     * 1: warnings but no errors
     * 0: all success
     */
    public int getReturnValue() {
        return returnValue;
    }


    @XmlRootElement(name = "overview")
    private static class XmlTestOverview {
        @XmlElementWrapper(name = "methods")
        @XmlElement(name = "method")
        private List<ApiTestMethodResult> details;

        @XmlElement(name = "version") private String version;
        @XmlElement(name = "environment") private String environment;

        @XmlElement(name = "TestUserUrn") private String testUserUrn;
        @XmlElement(name = "TestUserAuthorityUrn") private String testUserAuhUrn;
        @XmlElement(name = "TestedAuthorityUrn") private String testedAuthUrn;
        @XmlElement(name = "TestedClass") private String testedClass;
        @XmlElement(name = "TestedMethods") private String testedMethods;

        public XmlTestOverview() {
        }

        public XmlTestOverview(List<ApiTestMethodResult> resultList, String version, String environment) {
            this.version = version;
            this.environment = environment;
            details = new ArrayList<>(resultList);
        }
    }

    /**
     * write an XML file with an overview. This uses JAXB.
     * <p/>
     * Exceptions and API call details are not included: this requires adding a lot more JAXB annotations/code.
     * Only the method details are currently included, no general information about the test. To add that, don't use XmlTestOverview but annotate this class directly instead.
     */
    public void toXmlOverview(File file) {
        toXmlOverview(file, false);
    }

    public void toXmlOverview(File file, boolean showCredentials) {
        disableRemoveSecurityRisks = showCredentials; //a bit of a hack now, would be nice to do this cleaner

        try {
            XmlTestOverview x = new XmlTestOverview(resultList, getVersionString(), getEnvironment());
            x.testUserUrn = user.getUserUrnString();
            x.testUserAuhUrn = user.getUserAuthorityServer() == null ? "" : user.getUserAuthorityServer().getDefaultComponentManagerUrn();
            assert testedAuthority.getServerToConnect() != null;
            x.testedAuthUrn = testedAuthority.getServerToConnect().getDefaultComponentManagerUrn();
            x.testedClass = testClass.getName();
            if (testGroup != null)
                x.testedMethods = "Only group \"" + testGroup + "\" + dependencies";
            else
                x.testedMethods = "All";

            Marshaller m = JAXBContext.newInstance(XmlTestOverview.class).createMarshaller();
            m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
//            m.marshal(x, System.out);
            m.marshal(x, file);
        } catch (JAXBException e) {
            throw new RuntimeException("Error creating XML", e);
        }

        disableRemoveSecurityRisks = false;
    }

    public void toXmlOverview(Writer writer, boolean showCredentials) {
        disableRemoveSecurityRisks = showCredentials; //a bit of a hack now, would be nice to do this cleaner

        try {
            XmlTestOverview x = new XmlTestOverview(resultList, getVersionString(), getEnvironment());
            x.testUserUrn = user.getUserUrnString();
            x.testUserAuhUrn = user.getUserAuthorityServer() == null ? "" : user.getUserAuthorityServer().getDefaultComponentManagerUrn();
            assert testedAuthority.getServerToConnect() != null;
            x.testedAuthUrn = testedAuthority.getServerToConnect().getDefaultComponentManagerUrn();

            x.testedClass = testClass.getName();
            if (testGroup != null)
                x.testedMethods = "Only group \"" + testGroup + "\" + dependencies";
            else
                x.testedMethods = "All";

            Marshaller m = JAXBContext.newInstance(XmlTestOverview.class).createMarshaller();
            m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            m.marshal(x, writer);
        } catch (JAXBException e) {
            e.printStackTrace();
        }

        disableRemoveSecurityRisks = false;
    }

    @Override
    public void htmlContent(HtmlCanvas html) throws IOException {
        html.h1().content("Test Settings");
        html.div(class_("test-settings"));
        addKeyValue(html, "value-data", "Test User URN", user.getUserUrnString());
        addKeyValue(html, "value-data", "Test User Authority",
                user.getUserAuthorityServer() == null ? "unknown" : user.getUserAuthorityServer().getDefaultComponentManagerUrn());
        html.br();
        addKeyValue(html, "value-data", "Tested Aggregate Manager (if applicable)", testedAuthority.getServerToConnect().getDefaultComponentManagerUrn());
        html.br();
        addKeyValue(html, "value-data", "Test Class", testClass.getName());
        if (testGroup != null)
            addKeyValue(html, "value-nondata", "Tested Methods", "Only group \"" + testGroup + "\" + dependencies");
        else
            addKeyValue(html, "value-nondata", "Tested Methods", "All");
        try {
            addKeyValue(html, "value-data", "Test Description", ApiTest.getMetaData(testClass).getTestDescription());
        } catch (Exception e) {
            LOG.error("Error while retrieving Test Description", e);
            addKeyValue(html, "value-data", "Test Description", "Error while retrieving: " + e.getMessage());
        }

        html.br();
        addKeyValue(html, "value-data", "Tester Version", getVersionString());
        addKeyValue(html, "value-data", "Tester Environment", getEnvironment());

        html._div();

        html.h1().content("Overview");

        html.div(class_("times"));
        html.span(class_("timeheader")).content("Total duration ");
        html.span(class_("value-nondata")).content((getDurationMs() / 1000.0) + "s");
        html.span(class_("timeheader")).content(" from ");
        html.span(class_("value-nondata")).content(getStartDate() + "");
        html.span(class_("timeheader")).content(" to ");
        html.span(class_("value-nondata")).content(getStopDate() + "");
        html._div();

//            html.p().content("Test ran at: "+test.);

        int index = 0;
        for (ApiTestMethodResult res : resultList)
            writeOverviewItem(html, index++, res.state, res.methodName);

        index = 0;
        html.h1().content("Details");
        for (ApiTestMethodResult res : resultList) {
            writeGroup(html, index++, res.state, res.methodName, res.description,
                    res.timeMs, new Date(res.startTimeMs), new Date(res.stopTimeMs),
                    res.exception, res.apiCallDetails, res.getLogLines());
        }
    }






    //JAXB overhead
    public static class JAXBMutableApiTestMethodResult {
        private String methodName;
        private String description;
        private List<String> hardDependencies;
        private List<String> softDependencies;
        private List<LogEntry> logLines;
        private List<SerializableApiCallDetails> apiCallDetails;
        private Throwable exception;
        private TestResultState state;
        private long startTimeMs;
        private long stopTimeMs;
        private long timeMs;

        public JAXBMutableApiTestMethodResult() {
        }

        public JAXBMutableApiTestMethodResult(String methodName,
                                              String description,
                                              List<String> hardDependencies,
                                              List<String> softDependencies,
                                              List<LogEntry> logLines,
                                              List<SerializableApiCallDetails> apiCallDetails,
                                              Throwable exception,
                                              TestResultState state,
                                              long timeMs,
                                              long startTimeMs,
                                              long stopTimeMs) {
            this.methodName = methodName;
            this.description = description;
            this.hardDependencies = hardDependencies;
            this.softDependencies = softDependencies;
            this.logLines = logLines;
            this.apiCallDetails = apiCallDetails;
            this.exception = exception;
            this.state = state;
            this.startTimeMs = startTimeMs;
            this.stopTimeMs = stopTimeMs;
            this.timeMs = timeMs;
        }

        @XmlAttribute
        public String getMethodName() {
            return methodName;
        }

        public void setMethodName(String methodName) {
            this.methodName = methodName;
        }
        @XmlAttribute
        public String getDescription() {
            return description;
        }

        public void setDescription(String description) {
            this.description = description;
        }
        @XmlAttribute
        public List<String> getHardDependencies() {
            return hardDependencies;
        }

        public void setHardDependencies(List<String> hardDependencies) {
            this.hardDependencies = hardDependencies;
        }
        @XmlAttribute
        public List<String> getSoftDependencies() {
            return softDependencies;
        }

        public void setSoftDependencies(List<String> softDependencies) {
            this.softDependencies = softDependencies;
        }
        @XmlElementWrapper(name = "loglines")
        @XmlElement(name = "logline")
        public List<LogEntry> getLogLines() {
            return logLines;
        }

        public void setLogLines(List<LogEntry> logLines) {
            this.logLines = logLines;
        }
        @XmlElementWrapper(name = "apicalls")
        @XmlElement(name = "apicall")
        public List<SerializableApiCallDetails> getApiCallDetails() {
            return apiCallDetails;
        }

        public void setApiCallDetails(List<SerializableApiCallDetails> apiCallDetails) {
            this.apiCallDetails = apiCallDetails;
        }
        //        @XmlAttribute
        public Throwable getException() {
            return exception;
        }

        public void setException(Throwable exception) {
            this.exception = exception;
        }
        @XmlAttribute
        public TestResultState getState() {
            return state;
        }

        public void setState(TestResultState state) {
            this.state = state;
        }
        @XmlAttribute
        public long getStartTimeMs() {
            return startTimeMs;
        }

        public void setStartTimeMs(long startTimeMs) {
            this.startTimeMs = startTimeMs;
        }
        @XmlAttribute
        public long getStopTimeMs() {
            return stopTimeMs;
        }

        public void setStopTimeMs(long stopTimeMs) {
            this.stopTimeMs = stopTimeMs;
        }
        @XmlAttribute
        public long getTimeMs() {
            return timeMs;
        }

        public void setTimeMs(long timeMs) {
            this.timeMs = timeMs;
        }
    }

    public static class JAXBMutableApiTestMethodResultAdaptor extends XmlAdapter<JAXBMutableApiTestMethodResult, ApiTestMethodResult> {
        @Override
        public ApiTestMethodResult unmarshal(JAXBMutableApiTestMethodResult jaxbMutableApiTestMethodResult) throws Exception {
            return new ApiTestMethodResult(
                    jaxbMutableApiTestMethodResult.getMethodName(),
                    jaxbMutableApiTestMethodResult.getDescription(),
                    jaxbMutableApiTestMethodResult.getHardDependencies(),
                    jaxbMutableApiTestMethodResult.getSoftDependencies(),
                    jaxbMutableApiTestMethodResult.getLogLines(),
                    jaxbMutableApiTestMethodResult.getApiCallDetails(),
                    jaxbMutableApiTestMethodResult.getException(),
                    jaxbMutableApiTestMethodResult.getState(),
                    jaxbMutableApiTestMethodResult.getTimeMs(),
                    jaxbMutableApiTestMethodResult.getStartTimeMs(),
                    jaxbMutableApiTestMethodResult.getStopTimeMs()
            );
        }

        @Override
        public JAXBMutableApiTestMethodResult marshal(ApiTestMethodResult orig) throws Exception {
            return new JAXBMutableApiTestMethodResult(
                    orig.getMethodName(),
                    orig.getDescription(),
                    orig.getHardDependencies(),
                    orig.getSoftDependencies(),
                    orig.getLogLines(),
                    orig.getApiCallDetails(),
                    orig.getException(),
                    orig.getState(),
                    orig.getTimeMs(),
                    orig.getStartTimeMs(),
                    orig.getStopTimeMs()
            );
        }
    }


}
