package be.iminds.ilabt.jfed.util;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.*;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;

/**
 * This expects to be working on Maps containing other Maps, Lists, and primitives.
 * Keys are always String.
 * These are typically deserialized json or yaml.
 */
public class DefaultOmittingSerializationUtil {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(DefaultOmittingSerializationUtil.class);

    private DefaultOmittingSerializationUtil() {}

    private static void iterateSubMaps(Map<String, ?> source, BiConsumer<String, Map<String, ?>> action) {
        source.entrySet().stream()
                .filter(e -> e.getValue() instanceof Map)
                .forEach(e -> action.accept(e.getKey(), (Map<String, ?>) e.getValue()));
    }
    private static void iterateSubList(Map<String, ?> source, BiConsumer<String, List<?>> action) {
        source.entrySet().stream()
                .filter(e -> e.getValue() instanceof List)
                .forEach(e -> action.accept(e.getKey(), (List<?>) e.getValue()));
    }
    private static void iterateNonCollections(Map<String, ?> source, BiConsumer<String, Object> action) {
        source.entrySet().stream()
                .filter(e -> !(e.getValue() instanceof Map) && !(e.getValue() instanceof List))
                .forEach(e -> action.accept(e.getKey(), e.getValue()));
    }

    private static void iterateSubMaps(List<?> source, BiConsumer<Integer, Map<String, ?>> action) {
        for (int i = 0; i < source.size(); i++) {
            Object item = source.get(i);
            if (item instanceof Map)
                action.accept(i, (Map<String, ?>)item);
        }
    }
    private static void iterateSubList(List<?> source, BiConsumer<Integer, List<?>> action) {
        for (int i = 0; i < source.size(); i++) {
            Object item = source.get(i);
            if (item instanceof List)
                action.accept(i, (List<?>)item);
        }
    }
    private static void iterateNonCollections(List<?> source, BiConsumer<Integer, Object> action) {
        for (int i = 0; i < source.size(); i++) {
            Object item = source.get(i);
            if (!(item instanceof Map) && !(item instanceof List))
                action.accept(i, item);
        }
    }

    private static interface CopyJob {
        void run(Deque<CopyJob> jobs);
    }


    private static class CopyMapJob implements CopyJob {
        private final Map<String, ?> source;
        private final Map<String, Object> target;

        public CopyMapJob(Map<String, ?> source, Map<String,Object> target) {
            this.source = source;
            this.target = target;
        }

        public void run(Deque<CopyJob> jobs) {
            iterateNonCollections(source, (key, v) -> target.put(key, v));
            iterateSubMaps(source, (key, v) -> {
                Map copy = new HashMap(v);
                target.put(key, copy);
                jobs.push(new CopyMapJob(v, copy));
            });
            iterateSubList(source, (key, v) -> {
                List copy = new ArrayList<>(v);
                target.put(key, copy);
                jobs.push(new CopyListJob(v, copy));
            });
        }
    }
    private static class CopyListJob implements CopyJob {
        private final List<?> source;
        private final List<Object> target;

        public CopyListJob(List<?> source, List<Object> target) {
            this.source = source;
            this.target = target;
        }

        public void run(Deque<CopyJob> jobs) {
            iterateNonCollections(source, (index, v) -> target.set(index, v));
            iterateSubMaps(source, (index, v) -> {
                Map copy = new HashMap(v);
                target.set(index, copy);
                jobs.push(new CopyMapJob(v, copy));
            });
            iterateSubList(source, (index, v) -> {
                List copy = new ArrayList<>(v);
                target.set(index, copy);
                jobs.push(new CopyListJob(v, copy));
            });
        }
    }
    private static Map<String, Object> deepCopy(Map<String, ?> source) {
        Deque<CopyJob> jobs = new LinkedList<>();

        Map<String, Object> res = new HashMap<>();
        CopyMapJob job = new CopyMapJob(source, res);
        jobs.push(job);

        while (!jobs.isEmpty()) {
            jobs.pop().run(jobs);
        }

        return res;
    }

    private static class Keys {
        List<Object> keys;

        public Keys(List<Object> keys) {
            this.keys = new ArrayList<>(keys);
        }
        public Keys(Keys base, Object sub) {
            this.keys = new ArrayList<>(base.keys);
            this.keys.add(sub);
        }

        @Override
        public String toString() {
            return "Keys{" +
                    "" + keys.stream().map(Object::toString).collect(Collectors.joining(", ")) +
                    '}';
        }
    }
    private static class KeyIterateJob implements Runnable {
        private final Object collection;
        private final Keys curKeys;
        private final List<Keys> resultKeys;
        private final Collection<KeyIterateJob> jobs;

        public KeyIterateJob(Object collection, Keys curKeys, List<Keys> resultKeys, Collection<KeyIterateJob> jobs) {
            this.collection = collection;
            this.curKeys = curKeys;
            this.resultKeys = resultKeys;
            this.jobs = jobs;
        }

        @Override
        public void run() {
            if (collection instanceof Map) {
                Map<String, ?> m = (Map<String, ?>) collection;
                if (m.isEmpty()) {
                    resultKeys.add(curKeys);
                } else {
                    iterateNonCollections(m, (k, v) -> resultKeys.add(new Keys(curKeys, k)));
                    iterateSubMaps(m, (k, v) -> jobs.add(new KeyIterateJob(v, new Keys(curKeys, k), resultKeys, jobs)));
                    iterateSubList(m, (k, v) -> jobs.add(new KeyIterateJob(v, new Keys(curKeys, k), resultKeys, jobs)));
                }
            }
            if (collection instanceof List) {
                List<?> l = (List) collection;
                if (l.isEmpty()) {
                    resultKeys.add(curKeys);
                } else {
                    iterateNonCollections(l, (k, v) -> resultKeys.add(new Keys(curKeys, k)));
                    iterateSubMaps(l, (k, v) -> jobs.add(new KeyIterateJob(v, new Keys(curKeys, k), resultKeys, jobs)));
                    iterateSubList(l, (k, v) -> jobs.add(new KeyIterateJob(v, new Keys(curKeys, k), resultKeys, jobs)));
                }
            }
        }
    }
    private static List<Keys> getNestedKeys(Map<String, ?> source) {
        List<Keys> res = new ArrayList<>();

        Deque<KeyIterateJob> jobs = new LinkedList<>();

        KeyIterateJob job = new KeyIterateJob(source, new Keys(Collections.emptyList()), res, jobs);
        jobs.push(job);

        while (!jobs.isEmpty()) {
            jobs.pop().run();
        }

        return res;
    }

    @Nullable
    private static Object getByNestedKey(Map<String, ?> source, Keys keys) {
        LOG.debug("getByNestedKey(?, "+keys+")");
        Object curPos = source;
        for (Object k : keys.keys) {
            if (curPos instanceof Map) {
                curPos = ((Map)curPos).get((String)k);
            } else {
                if (curPos instanceof List) {
                    int index = (Integer)k;
                    List l = (List)curPos;
                    if (index >= l.size())
                        return null;
                    else
                        curPos = l.get((Integer)k);
                } else
                    return null;
            }
        }
        return curPos;
    }

    private static void removeByKey(Map<String, ?> target, Keys keys) {
        LOG.debug("removeByKey(?, "+keys+")");
        Object curPos = target;
        assert keys.keys.size() > 0;
        List<Object> firstKeys = new LinkedList<>(keys.keys);
        Object lastKey = ((LinkedList<Object>) firstKeys).removeLast();
        for (Object k : firstKeys) {
            if (curPos instanceof Map) {
                curPos = ((Map)curPos).get((String)k);
            } else {
                if (curPos instanceof List) {
                    int index = (Integer)k;
                    List l = (List)curPos;
                    if (index >= l.size())
                        return;
                    else
                        curPos = l.get(index);
                } else
                    return;
            }
        }
        if (curPos instanceof Map) {
            ((Map)curPos).remove((String)lastKey);
        } else {
            if (curPos instanceof List) {
                int index = (Integer)lastKey;
                List l = (List)curPos;
                if (index < l.size())
                    l.remove(index);
            }
        }
    }

    /**
     * removes empty maps or lists from the "tree" leading up to this key
     *
     * @param target
     * @param keys
     * @return true if an (empty) map or list was removed
     */
    private static boolean removeByKeyIfEmpty(Map<String, ?> target, Keys keys) {
        LOG.debug("removeByKeyIfEmpty(?, "+keys+")");
        Object curPos = target;
        assert keys.keys.size() > 0;
        for (Object k : keys.keys) {
            if (curPos instanceof Map) {
                Map m = (Map)curPos;
                Object newCurPos = m.get((String)k);
                if (newCurPos instanceof Map && ((Map)newCurPos).isEmpty()) {
                    m.remove(k);
                    LOG.debug("removeByKeyIfEmpty removes "+k);
                    return true;
                }
                if (newCurPos instanceof List && ((List)newCurPos).isEmpty()) {
                    m.remove(k);
                    LOG.debug("removeByKeyIfEmpty removes "+k);
                    return true;
                }
                curPos = newCurPos;
            } else {
                if (curPos instanceof List) {
                    int index = (Integer)k;
                    List l = (List)curPos;
                    if (index >= l.size())
                        return false;
                    else {
                        Object newCurPos = l.get(index);
                        if (newCurPos instanceof Map && ((Map)newCurPos).isEmpty()) {
                            l.remove(index);
                            LOG.debug("removeByKeyIfEmpty removes "+index);
                            return true;
                        }
                        if (newCurPos instanceof List && ((List)newCurPos).isEmpty()) {
                            l.remove(index);
                            LOG.debug("removeByKeyIfEmpty removes "+index);
                            return true;
                        }
                        curPos = newCurPos;
                    }
                } else
                    return false;
            }
        }

        return false;
    }

    public static Map<String, ?> omitDefault(Map<String, ?> target, Map<String, ?> defaults) {
        Map<String, ?> res = deepCopy(target);
        LOG.debug("target="+debugPrint(target));
        LOG.debug("defaults="+debugPrint(defaults));

        if (!debugPrint(target).equals(debugPrint(res))) {
            LOG.error("deepCopy differs!");
            LOG.debug("target="+debugPrint(target));
            LOG.debug("deepCopy="+debugPrint(res));
        }

        List<Keys> keysList = Collections.unmodifiableList(getNestedKeys(target));
        LOG.debug("keysList="+keysList);

        List<Keys> keysToRemove = new ArrayList<>();
        for (Keys keys: keysList) {
            Object targetVal = getByNestedKey(target, keys);
            Object defaultVal = getByNestedKey(defaults, keys);
            if (defaultVal == targetVal)
                keysToRemove.add(keys);
        }

        keysToRemove = Collections.unmodifiableList(keysToRemove);
        LOG.debug("keysToRemove="+keysToRemove);
        for (Keys keys: keysToRemove) {
            removeByKey(res, keys);
        }

        boolean hasChanged = true;
        while (hasChanged) {
            hasChanged = false;
            for (Keys keys : keysList) {
                hasChanged |= removeByKeyIfEmpty(res, keys);
            }
            LOG.debug("removeByKeyIfEmpty hasChanged="+hasChanged);
        }

        LOG.debug("res="+debugPrint(res));
        return res;
    }


    private static String debugPrint(Object o) {
        try {
            return new ObjectMapper()
                    .writerWithDefaultPrettyPrinter()
                    .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                    .writeValueAsString(o);
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }
    }
}
