package be.iminds.ilabt.jfed.experiment.setup;

import be.iminds.ilabt.jfed.rspec.AdvertisementBasedRspecGeneratorConfig;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;

public class ExperimentSetupConfig {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSetupConfig.class);

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Nonnull protected final Slice slice;
    @Nullable protected final RequestRSpec requestRSpec;
    @Nullable protected final ESpec eSpec;
    @Nonnull protected final Provision provision;
    @Nonnull protected final WaitForReady waitForReady;
    protected final boolean overrideESpecRSpec;
    protected final boolean runLinkTest;

    public ExperimentSetupConfig() {
        this.slice = new Slice();
        this.requestRSpec = RequestRSpec.getDefault();
        this.eSpec = null;
        this.provision = new Provision();
        this.waitForReady = new WaitForReady();
        this.overrideESpecRSpec = false;
        this.runLinkTest = false;
    }

    @JsonCreator
    public ExperimentSetupConfig(@Nullable @JsonProperty("slice") Slice slice,
                    @Nullable @JsonProperty("requestRSpec") RequestRSpec requestRSpec,
                    @Nullable @JsonProperty("eSpec") ESpec eSpec,
                    @Nullable @JsonProperty("provision") Provision provision,
                    @Nullable @JsonProperty("waitForReady") WaitForReady waitForReady,
                    @Nullable @JsonProperty("overrideESpecRSpec") Boolean overrideESpecRSpec,
                    @Nullable @JsonProperty("runLinkTest") Boolean runLinkTest) {
        this.slice = slice == null ? new ExperimentSetupConfig().getSlice() : slice;
        this.requestRSpec = requestRSpec == null ?
                (eSpec == null ? new ExperimentSetupConfig().getRequestRSpec() : null) :
                requestRSpec;
        this.eSpec = eSpec == null ? null : eSpec;
        this.provision = provision == null ? new ExperimentSetupConfig().getProvision() : provision;
        this.waitForReady = waitForReady == null ? new ExperimentSetupConfig().getWaitForReady() : waitForReady;
        this.overrideESpecRSpec = overrideESpecRSpec == null ? new ExperimentSetupConfig().getOverrideESpecRSpec() : overrideESpecRSpec;
        this.runLinkTest = runLinkTest == null ? new ExperimentSetupConfig().getRunLinkTest() : runLinkTest;
    }

    @JsonIgnore
    public boolean isValid() {
        boolean res = getSlice().isValid();
        RequestRSpec requestRspec = getRequestRSpec();
        ESpec espec = getESpec();
        res = res && (requestRspec == null || requestRspec.isValid());
        res = res && (espec == null || espec.isValid());
        res = res && (espec != null || requestRspec != null);
        res = res && getProvision().isValid();
        res = res && getWaitForReady().isValid();
        if (!getProvision().isEnabled() && !getWaitForReady().isEnabled())
            throw new IllegalArgumentException("Config Error: if the Provision step is disabled, waitForReady should be disabled too.");
        if (getOverrideESpecRSpec()) {
            if (espec == null || requestRspec == null)
                throw new IllegalArgumentException("Config Error: overrideESpecRSpec=true requires specifying both an RSpec and an ESpec");
        } else {
            if (espec != null && requestRspec != null)
                throw new IllegalArgumentException("Config Error: Specifying both an RSpec and an ESpec requires overrideESpecRSpec=true");
        }
        return res;
    }

    @JsonProperty
    @Nonnull
    public Slice getSlice() {
        return slice;
    }

    @JsonProperty
    @Nullable
    public RequestRSpec getRequestRSpec() {
        return requestRSpec;
    }

    @JsonProperty("eSpec")
    @Nullable
    public ESpec getESpec() {
        return eSpec;
    }

    @JsonProperty
    @Nonnull
    public Provision getProvision() {
        return provision;
    }

    @JsonProperty
    @Nonnull
    public WaitForReady getWaitForReady() {
        return waitForReady;
    }


    @JsonProperty("overrideESpecRSpec")
    public boolean getOverrideESpecRSpec() {
        return overrideESpecRSpec;
    }

    @JsonProperty("runLinkTest")
    public boolean getRunLinkTest() {
        return runLinkTest;
    }


    public enum RequestRSpecSource { PROVIDE_URL, PROVIDE_FILE, PROVIDE_CONTENT,
        GENERATE_USING_JFED, GENERATE_SIMPLE, GENERATE_USING_ADVERTISEMENT,
        ALREADY_PROVISIONED;

        public boolean isProvide() {
            return this == PROVIDE_URL ||
                    this == PROVIDE_FILE ||
                    this == PROVIDE_CONTENT;
        }
    }

    @JsonTypeInfo(
            use = JsonTypeInfo.Id.NAME,
            include = JsonTypeInfo.As.PROPERTY,
            property = "source",
            defaultImpl = GenerateUsingJFedRequestRSpec.class,
            visible = true)
    @JsonSubTypes({
            @JsonSubTypes.Type(value = GenerateSimpleRequestRSpec.class, name = "GENERATE_SIMPLE"),
            @JsonSubTypes.Type(value = ProvidedUrlRequestRSpec.class, name = "PROVIDE_URL"),
            @JsonSubTypes.Type(value = ProvidedFileRequestRSpec.class, name = "PROVIDE_FILE"),
            @JsonSubTypes.Type(value = ProvidedContentRequestRSpec.class, name = "PROVIDE_CONTENT"),
            @JsonSubTypes.Type(value = GenerateUsingJFedRequestRSpec.class, name = "GENERATE_USING_JFED"),
            @JsonSubTypes.Type(value = GenerateUsingAdvertisementRequestRSpec.class, name = "GENERATE_USING_ADVERTISEMENT"),
            @JsonSubTypes.Type(value = AlreadyProvisionedRequestRSpec.class, name = "ALREADY_PROVISIONED")
    })
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static abstract class RequestRSpec {
        @Nonnull
        protected final RequestRSpecSource source;

        //used for all methods
        private final boolean autoAssignSpecificNodesUsingListResources;
        private final boolean useNitosLease;

        @JsonCreator
        public RequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                            @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                            @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            assert source != null;
            this.source = source;
            this.autoAssignSpecificNodesUsingListResources =
                    autoAssignSpecificNodesUsingListResources == null ?
                            getDefault().getAutoAssignSpecificNodesUsingListResources() :
                            autoAssignSpecificNodesUsingListResources;
            this.useNitosLease = useNitosLease == null ? getDefault().getUseNitosLease() : useNitosLease;
        }

        public static RequestRSpec getDefault() {
            return new GenerateUsingJFedRequestRSpec();
        }


        @JsonCreator
        public static RequestRSpec createRequestRSpecFromJsonObject(@Nonnull Map<String,Object> props) {
            //TODO: Check if types are correct etc
            RequestRSpecSource source =
                    props.get("source") == null ?
                            null :
                            RequestRSpecSource.valueOf((String) props.get("source"));
            Boolean autoAssignSpecificNodesUsingListResources = (Boolean) props.get("autoAssignSpecificNodesUsingListResources");
            Boolean useNitosLease = (Boolean) props.get("useNitosLease");

            LOG.debug("createRequestRSpecFromJsonObject with source="+source);

            if (source == null) {
                LOG.debug("createRequestRSpecFromJsonObject with source=null -> reverting all to defaults");
                return getDefault();
            }

            switch (source) {
                case ALREADY_PROVISIONED: {
                    return new AlreadyProvisionedRequestRSpec(
                            source,
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                case PROVIDE_FILE: {
                    return new ProvidedFileRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            (String) props.get("bindUnboundNodesUrn"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case PROVIDE_URL: {
                    return new ProvidedUrlRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            (String) props.get("bindUnboundNodesUrn"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case PROVIDE_CONTENT: {
                    return new ProvidedContentRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            (String) props.get("bindUnboundNodesUrn"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case GENERATE_USING_ADVERTISEMENT: {
                    AdvertisementBasedRspecGeneratorConfig advertisementBasedRspecGeneratorConfig =
                            MAPPER.convertValue(
                                    props.get("config"),
                                    AdvertisementBasedRspecGeneratorConfig.class);
                    return new GenerateUsingAdvertisementRequestRSpec(source,
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease,
                            advertisementBasedRspecGeneratorConfig);
                }
                case GENERATE_SIMPLE: {
                    return new GenerateSimpleRequestRSpec(source,
                            (Integer) props.get("nodeCount"),
                            (String) props.get("nodeSliverType"),
                            (Boolean) props.get("nodeExclusive"),
                            (List<String>) props.get("nodeComponentUrn"),
                            (List<String>) props.get("nodeAnsibleGroup"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                case GENERATE_USING_JFED: {
                    return new GenerateUsingJFedRequestRSpec(source,
                            (Integer) props.get("nodeCount"),
                            (String) props.get("jFedResourceClass"),
                            (List<String>) props.get("nodeAnsibleGroup"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                default:
                    throw new IllegalArgumentException("Config Error: unsupported rspec.source == "+source);
            }
        }

        @JsonIgnore
        public abstract boolean isValid();

        @JsonProperty
        @Nonnull
        public RequestRSpecSource getSource() {
            return source;
        }

        @JsonProperty
        public boolean getAutoAssignSpecificNodesUsingListResources() {
            return autoAssignSpecificNodesUsingListResources;
        }

        @JsonProperty
        public boolean getUseNitosLease() {
            return useNitosLease;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof RequestRSpec)) return false;

            RequestRSpec that = (RequestRSpec) o;

            if (autoAssignSpecificNodesUsingListResources != that.autoAssignSpecificNodesUsingListResources)
                return false;
            if (useNitosLease != that.useNitosLease) return false;
            if (source != that.source) return false;
            return true;
        }

        @Override
        public int hashCode() {
            int result = source.hashCode();
            result = 31 * result + (autoAssignSpecificNodesUsingListResources ? 1 : 0);
            result = 31 * result + (useNitosLease ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateUsingJFedRequestRSpec extends RequestRSpec {
        private final int nodeCount;
        @Nullable private final String jFedResourceClass; //if null, a class will be picked automatically, using the order in the jFed GUI (top being highest priority)
        @Nonnull private final List<String> nodeAnsibleGroup;

        public GenerateUsingJFedRequestRSpec() {
            super(RequestRSpecSource.GENERATE_USING_JFED, false, false);
            this.nodeCount = 1;
            this.jFedResourceClass = null;
            this.nodeAnsibleGroup = Collections.emptyList();
        }

        @JsonCreator
        public GenerateUsingJFedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                             @Nullable @JsonProperty("nodeCount") Integer nodeCount,
                                             @Nullable @JsonProperty("jFedResourceClass") String jFedResourceClass,
                                             @Nullable @JsonProperty("nodeAnsibleGroup") List<String> nodeAnsibleGroup,
                                             @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                             @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(RequestRSpecSource.GENERATE_USING_JFED, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_USING_JFED;
            assert this.getSource() == RequestRSpecSource.GENERATE_USING_JFED;
            this.nodeCount = nodeCount == null ? new GenerateUsingJFedRequestRSpec().getNodeCount() : nodeCount;
            this.jFedResourceClass = jFedResourceClass == null ? new GenerateUsingJFedRequestRSpec().getjFedResourceClass() : jFedResourceClass;
            this.nodeAnsibleGroup = nodeAnsibleGroup == null ? new GenerateSimpleRequestRSpec().getNodeAnsibleGroup() : nodeAnsibleGroup;
        }

        @JsonIgnore
        public boolean isValid() {
            if (nodeCount <= 0)
                throw new IllegalArgumentException("Config Error: rspec.nodeCount == "+nodeCount+" makes no sense if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        public int getNodeCount() {
            return nodeCount;
        }

        @JsonProperty
        @Nullable
        public String getjFedResourceClass() {
            return jFedResourceClass;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeAnsibleGroup() {
            return nodeAnsibleGroup;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateUsingJFedRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateUsingJFedRequestRSpec that = (GenerateUsingJFedRequestRSpec) o;

            if (nodeCount != that.nodeCount) return false;
            if (!nodeAnsibleGroup.equals(that.nodeAnsibleGroup)) return false;
            return jFedResourceClass != null ? jFedResourceClass.equals(that.jFedResourceClass) : that.jFedResourceClass == null;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + nodeCount;
            result = 31 * result + (jFedResourceClass != null ? jFedResourceClass.hashCode() : 0);
            result = 31 * result + nodeAnsibleGroup.hashCode();
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class AlreadyProvisionedRequestRSpec extends RequestRSpec {
        public AlreadyProvisionedRequestRSpec() {
            super(RequestRSpecSource.ALREADY_PROVISIONED,
                    false,
                    false);
        }

        @JsonCreator
        public AlreadyProvisionedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                              @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                              @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.ALREADY_PROVISIONED;
            assert this.getSource() == RequestRSpecSource.ALREADY_PROVISIONED;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedContentRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedContentRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_CONTENT,
                    null,
                    false,
                    null,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedContentRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                           @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                           @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                           @Nullable @JsonProperty("bindUnboundNodesUrn") String bindUnboundNodesUrn,
                                           @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                           @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes, bindUnboundNodesUrn,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_CONTENT;
            assert this.getSource() == RequestRSpecSource.PROVIDE_CONTENT;
        }
    }
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedUrlRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedUrlRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_URL,
                    null,
                    false,
                    null,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedUrlRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                       @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                       @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                       @Nullable @JsonProperty("bindUnboundNodesUrn") String bindUnboundNodesUrn,
                                       @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                       @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes, bindUnboundNodesUrn,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_URL;
            assert this.getSource() == RequestRSpecSource.PROVIDE_URL;
        }
    }
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedFileRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedFileRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_FILE,
                    null,
                    false,
                    null,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedFileRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                        @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                        @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                        @Nullable @JsonProperty("bindUnboundNodesUrn") String bindUnboundNodesUrn,
                                        @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                        @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes, bindUnboundNodesUrn,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_FILE;
            assert this.getSource() == RequestRSpecSource.PROVIDE_FILE;
        }
    }

    @JsonTypeInfo(
            use = JsonTypeInfo.Id.NAME,
            include = JsonTypeInfo.As.PROPERTY,
            property = "source",
            defaultImpl = ProvidedContentRequestRSpec.class,
            visible = true)
    @JsonSubTypes({
            @JsonSubTypes.Type(value = ProvidedUrlRequestRSpec.class, name = "PROVIDE_URL"),
            @JsonSubTypes.Type(value = ProvidedFileRequestRSpec.class, name = "PROVIDE_FILE"),
            @JsonSubTypes.Type(value = ProvidedContentRequestRSpec.class, name = "PROVIDE_CONTENT")
    })
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static abstract class ProvidedRequestRSpec extends RequestRSpec {
        @Nullable private final String providedContentSource; //url, file or content as specified in source

        private final boolean bindTestedServerToUnboundNodes;
        @Nullable private final String bindUnboundNodesUrn;

        public ProvidedRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_CONTENT,
                    false,
                    false);
            this.providedContentSource = "";
            //best that bindTestedServerToUnboundNodes is default. For geni tutorial tests, this is very convenient.
            //Also, jFed Gui Logic doesn't work correctly for unbound nodes this anyway.
            this.bindTestedServerToUnboundNodes = true;
            this.bindUnboundNodesUrn = null;
        }

        @JsonCreator
        public ProvidedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                    @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                    @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                    @Nullable @JsonProperty("bindUnboundNodesUrn") String bindUnboundNodesUrn,
                                    @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                    @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_CONTENT ||
                    source == RequestRSpecSource.PROVIDE_FILE ||
                    source == RequestRSpecSource.PROVIDE_URL;
            assert this.getSource() == RequestRSpecSource.PROVIDE_CONTENT ||
                    this.getSource() == RequestRSpecSource.PROVIDE_FILE ||
                    this.getSource() == RequestRSpecSource.PROVIDE_URL;
            this.providedContentSource = providedContentSource;
            this.bindTestedServerToUnboundNodes = bindTestedServerToUnboundNodes == null ?
                    new ProvidedContentRequestRSpec().getBindTestedServerToUnboundNodes() :
                    bindTestedServerToUnboundNodes;
            this.bindUnboundNodesUrn = bindUnboundNodesUrn;
        }

        @JsonIgnore
        public boolean isValid() {
            if (providedContentSource == null)
                throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be null if rspec.source == "+source);
            if (providedContentSource.trim().isEmpty())
                throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be empty if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        @Nullable
        public String getProvidedContentSource() {
            return providedContentSource;
        }

        /**
         * This option makes only sense for automated tests.
         *
         * If this is true, it implies "bindUnboundNodesUrn" is "auto".
         * (so bindUnboundNodesUrn will be ignored)
         *
         * @return
         */
        @JsonProperty
        public boolean getBindTestedServerToUnboundNodes() {
            return bindTestedServerToUnboundNodes;
        }

        /**
         * Urn of the AM to bind unbound nodes to.
         * If null, unbound nodes should not be bound.
         * @return
         */
        @Nullable
        @JsonProperty
        public String getBindUnboundNodesUrn() {
            return bindUnboundNodesUrn;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof ProvidedRequestRSpec)) return false;
            if (!super.equals(o)) return false;
            ProvidedRequestRSpec that = (ProvidedRequestRSpec) o;
            return bindTestedServerToUnboundNodes == that.bindTestedServerToUnboundNodes &&
                    Objects.equals(providedContentSource, that.providedContentSource) &&
                    Objects.equals(bindUnboundNodesUrn, that.bindUnboundNodesUrn);
        }

        @Override
        public int hashCode() {
            return Objects.hash(super.hashCode(), providedContentSource, bindTestedServerToUnboundNodes, bindUnboundNodesUrn);
        }

        @Override
        public String toString() {
            return "ProvidedRequestRSpec{" +
                    "source='" + getSource() + '\'' +
                    ", providedContentSource=" + providedContentSource +
                    ", bindTestedServerToUnboundNodes=" + bindTestedServerToUnboundNodes +
                    ", bindUnboundNodesUrn=" + bindUnboundNodesUrn +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateUsingAdvertisementRequestRSpec extends RequestRSpec {
        @Nonnull private final AdvertisementBasedRspecGeneratorConfig
                config;

        public GenerateUsingAdvertisementRequestRSpec() {
            super(RequestRSpecSource.GENERATE_USING_ADVERTISEMENT, false, false);
            this.config = new AdvertisementBasedRspecGeneratorConfig(Collections.singletonList(
                    new AdvertisementBasedRspecGeneratorConfig.NodeInfo(
                            1, null,"node", null,
                            null, null, null,
                            null, null, null,
                            null, null, true, null
                    )
            ));
        }

        @JsonCreator
        public GenerateUsingAdvertisementRequestRSpec(
                @Nonnull @JsonProperty("source") RequestRSpecSource source,
                @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease,
                @Nullable @JsonProperty("config") AdvertisementBasedRspecGeneratorConfig config) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_USING_ADVERTISEMENT;
            assert this.getSource() == RequestRSpecSource.GENERATE_USING_ADVERTISEMENT;
            this.config =
                    config == null ?
                            new GenerateUsingAdvertisementRequestRSpec().getConfig() :
                            config;
        }

        @JsonIgnore
        public boolean isValid() {
            if (config == null)
                throw new IllegalArgumentException("Config Error: rspec.config required for rspec.source == "+source);
            if (config.getNodes().isEmpty())
                throw new IllegalArgumentException("Config Error: rspec.config.nodes may not be empty for rspec.source == "+source);
            return true;
        }

        @JsonProperty
        @Nonnull
        public AdvertisementBasedRspecGeneratorConfig getConfig() {
            return config;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateUsingAdvertisementRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateUsingAdvertisementRequestRSpec that = (GenerateUsingAdvertisementRequestRSpec) o;

            //noinspection ConstantConditions
            return config != null ? config.equals(that.config) : that.config == null;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            //noinspection ConstantConditions
            result = 31 * result + (config != null ? config.hashCode() : 0);
            return result;
        }

        @Override
        public String toString() {
            return "GenerateUsingAdvertisementRequestRSpec{" +
                    "source=" + this.getSource() +
                    ", autoAssignSpecificNodesUsingListResources=" + this.getAutoAssignSpecificNodesUsingListResources() +
                    ", useNitosLease=" + this.getUseNitosLease() +
                    ", config=" + config +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateSimpleRequestRSpec extends RequestRSpec {
        private final int nodeCount;
        @Nonnull private final String nodeSliverType;
        private final boolean nodeExclusive;
        @Nonnull private final List<String> nodeComponentUrn;
        @Nonnull private final List<String> nodeAnsibleGroup;

        public GenerateSimpleRequestRSpec() {
            super(RequestRSpecSource.GENERATE_SIMPLE, false, false);
            this.nodeCount = 1;
            this.nodeSliverType = "raw-pc";
            this.nodeExclusive = true;
            this.nodeComponentUrn = Collections.emptyList();
            this.nodeAnsibleGroup = Collections.emptyList();
        }

        @JsonCreator
        public GenerateSimpleRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                          @Nullable @JsonProperty("nodeCount") Integer nodeCount,
                                          @Nullable @JsonProperty("nodeSliverType") String nodeSliverType,
                                          @Nullable @JsonProperty("nodeExclusive") Boolean nodeExclusive,
                                          @Nullable @JsonProperty("nodeComponentUrn") List<String> nodeComponentUrn,
                                          @Nullable @JsonProperty("nodeAnsibleGroup") List<String> nodeAnsibleGroup,
                                          @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                          @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_SIMPLE;
            assert this.getSource() == RequestRSpecSource.GENERATE_SIMPLE;

            this.nodeCount = nodeCount == null ? new GenerateSimpleRequestRSpec().getNodeCount() : nodeCount;
            this.nodeSliverType = nodeSliverType == null ? new GenerateSimpleRequestRSpec().getNodeSliverType() : nodeSliverType;
            this.nodeExclusive = nodeExclusive == null ? new GenerateSimpleRequestRSpec().getNodeExclusive() : nodeExclusive;
            this.nodeComponentUrn = nodeComponentUrn == null ? new GenerateSimpleRequestRSpec().getNodeComponentUrn() : nodeComponentUrn;
            this.nodeAnsibleGroup = nodeAnsibleGroup == null ? new GenerateSimpleRequestRSpec().getNodeAnsibleGroup() : nodeAnsibleGroup;
        }

        @JsonIgnore
        public boolean isValid() {
            if (nodeCount <= 0)
                throw new IllegalArgumentException("Config Error: rspec.nodeCount == "+nodeCount+" makes no sense if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        public int getNodeCount() {
            return nodeCount;
        }

        @JsonProperty
        @Nonnull
        public String getNodeSliverType() {
            return nodeSliverType;
        }

        @JsonProperty
        public boolean getNodeExclusive() {
            return nodeExclusive;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeComponentUrn() {
            return nodeComponentUrn;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeAnsibleGroup() {
            return nodeAnsibleGroup;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateSimpleRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateSimpleRequestRSpec that = (GenerateSimpleRequestRSpec) o;

            if (nodeCount != that.nodeCount) return false;
            if (nodeExclusive != that.nodeExclusive) return false;
            if (!nodeSliverType.equals(that.nodeSliverType)) return false;
            if (!nodeAnsibleGroup.equals(that.nodeAnsibleGroup)) return false;
            return nodeComponentUrn.equals(that.nodeComponentUrn);
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + nodeCount;
            result = 31 * result + nodeSliverType.hashCode();
            result = 31 * result + (nodeExclusive ? 1 : 0);
            result = 31 * result + nodeComponentUrn.hashCode();
            result = 31 * result + nodeAnsibleGroup.hashCode();
            return result;
        }
    }

    public enum ProjectSource { NONE, AUTO, PROVIDED }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Slice {
        @Nullable private final String sliceName;
        private final boolean failOnExistingSlice;
        private final boolean failOnNonExistingSlice;
        private final long expireTimeMin;
        private final boolean renewExistingSliceIfNeeded;

        @Nonnull private final ProjectSource projectSource;
        private final boolean failIfNoProject; //if projectSource is AUTO or PROVIDED
        @Nullable private final String project; //if projectSource is PROVIDED

        public Slice() {
            this.sliceName = null;
            this.failOnExistingSlice = true;
            this.failOnNonExistingSlice = false;
            this.expireTimeMin = 60;
            this.renewExistingSliceIfNeeded = true;
            this.projectSource = ProjectSource.AUTO;
            this.failIfNoProject = false;
            this.project = null;
        }

        @JsonCreator
        public Slice(@Nullable @JsonProperty("sliceName") String sliceName,
                     @Nullable @JsonProperty("failOnExistingSlice") Boolean failOnExistingSlice,
                     @Nullable @JsonProperty("failOnNonExistingSlice") Boolean failOnNonExistingSlice,
                     @Nullable @JsonProperty("expireTimeMin") Long expireTimeMin,
                     @Nullable @JsonProperty("renewExistingSliceIfNeeded") Boolean renewExistingSliceIfNeeded,
                     @Nullable @JsonProperty("projectSource") ProjectSource projectSource,
                     @Nullable @JsonProperty("failIfNoProject") Boolean failIfNoProject,
                     @Nullable @JsonProperty("project") String project) {
            this.sliceName = sliceName == null ? new Slice().getSliceName() : sliceName;
            this.failOnExistingSlice = failOnExistingSlice == null ? new Slice().getFailOnExistingSlice() : failOnExistingSlice;
            this.failOnNonExistingSlice = failOnNonExistingSlice == null ? new Slice().getFailOnNonExistingSlice() : failOnNonExistingSlice;
            this.expireTimeMin = expireTimeMin == null ? new Slice().getExpireTimeMin() : expireTimeMin;
            this.renewExistingSliceIfNeeded = renewExistingSliceIfNeeded == null ? new Slice().getRenewExistingSliceIfNeeded() : renewExistingSliceIfNeeded;
            this.projectSource = projectSource == null ? new Slice().getProjectSource() : projectSource;
            this.failIfNoProject = failIfNoProject == null ? new Slice().getFailIfNoProject() : failIfNoProject;
            this.project = project == null ? new Slice().getProject() : project;
        }

        @JsonIgnore
        public boolean isValid() {
            if (expireTimeMin <= 0)
                throw new IllegalArgumentException("Config Error: slice.expireTimeMin <= makes no sense: "+expireTimeMin);
            if (projectSource == ProjectSource.PROVIDED && project == null && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and project==null and projectSource == ProjectSource.PROVIDED will always fail");
            if (projectSource == ProjectSource.NONE && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and projectSource == ProjectSource.NONE will always fail");
            if (failOnExistingSlice && failOnNonExistingSlice)
                throw new IllegalArgumentException("Config Error: failOnExistingSlice and failOnNonExistingSlice should not both be true");

            return true;
        }

        @Nullable
        public String getSliceName() {
            return sliceName;
        }

        public boolean getFailOnExistingSlice() {
            return failOnExistingSlice;
        }

        public boolean getFailOnNonExistingSlice() {
            return failOnNonExistingSlice;
        }

        public long getExpireTimeMin() {
            return expireTimeMin;
        }

        public boolean getRenewExistingSliceIfNeeded() {
            return renewExistingSliceIfNeeded;
        }

        @Nonnull
        public ProjectSource getProjectSource() {
            return projectSource;
        }

        public boolean getFailIfNoProject() {
            return failIfNoProject;
        }

        @Nullable
        public String getProject() {
            return project;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Slice)) return false;

            Slice slice = (Slice) o;

            if (failOnExistingSlice != slice.failOnExistingSlice) return false;
            if (failOnNonExistingSlice != slice.failOnNonExistingSlice) return false;
            if (expireTimeMin != slice.expireTimeMin) return false;
            if (renewExistingSliceIfNeeded != slice.renewExistingSliceIfNeeded) return false;
            if (failIfNoProject != slice.failIfNoProject) return false;
            if (sliceName != null ? !sliceName.equals(slice.sliceName) : slice.sliceName != null) return false;
            if (projectSource != slice.projectSource) return false;
            return project != null ? project.equals(slice.project) : slice.project == null;
        }

        @Override
        public int hashCode() {
            int result = sliceName != null ? sliceName.hashCode() : 0;
            result = 31 * result + (failOnExistingSlice ? 1 : 0);
            result = 31 * result + (failOnNonExistingSlice ? 1 : 0);
            result = 31 * result + (int) (expireTimeMin ^ (expireTimeMin >>> 32));
            result = 31 * result + (renewExistingSliceIfNeeded ? 1 : 0);
            result = 31 * result + projectSource.hashCode();
            result = 31 * result + (failIfNoProject ? 1 : 0);
            result = 31 * result + (project != null ? project.hashCode() : 0);
            return result;
        }
    }


    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ESpecLogStorageDetails {
        @Nullable private final String path; //if null, match all logs
        @Nullable private final String nodeClientId; //if null, match all nodes
        private final boolean storeInFile;
        private final boolean storeInLog;

        public ESpecLogStorageDetails() {
            this.path = null;
            this.nodeClientId = null;
            this.storeInFile = false;
            this.storeInLog = false;
        }

        @JsonCreator
        public ESpecLogStorageDetails(@Nullable @JsonProperty("path") String path,
                                      @Nullable @JsonProperty("nodeClientId") String nodeClientId,
                                      @JsonProperty("storeInFile") boolean storeInFile,
                                      @JsonProperty("storeInLog") boolean storeInLog) {
            this.path = path;
            this.nodeClientId = nodeClientId;
            this.storeInFile = storeInFile;
            this.storeInLog = storeInLog;
        }

        @JsonProperty
        @Nullable
        public String getPath() {
            return path;
        }

        @JsonProperty
        @Nullable
        public String getNodeClientId() {
            return nodeClientId;
        }

        @JsonProperty
        public boolean isStoreInFile() {
            return storeInFile;
        }

        @JsonProperty
        public boolean isStoreInLog() {
            return storeInLog;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof ESpecLogStorageDetails)) return false;
            ESpecLogStorageDetails that = (ESpecLogStorageDetails) o;
            return storeInFile == that.storeInFile &&
                    storeInLog == that.storeInLog &&
                    Objects.equals(path, that.path) &&
                    Objects.equals(nodeClientId, that.nodeClientId);
        }

        @Override
        public int hashCode() {
            return Objects.hash(path, nodeClientId, storeInFile, storeInLog);
        }
    }

    public enum ESpecSource {PROVIDE_ARCHIVE_URL, PROVIDE_ARCHIVE_FILE, PROVIDE_DIR, PROVIDE_GIT_REPO_DIR, DIRECT };
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ESpec {
        @Nonnull private final ESpecSource source;
        @Nullable private final String providedContentSource; //url, file or content as specified in source

        @Nullable private final List<ESpecLogStorageDetails> logStorage;

        public ESpec() {
            this.source = ESpecSource.PROVIDE_DIR;
            this.providedContentSource = null;
            this.logStorage = null;
        }

        @JsonCreator
        public ESpec(@Nullable @JsonProperty("source") ESpecSource source,
                     @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                     @Nullable @JsonProperty("logStorage") List<ESpecLogStorageDetails> logStorage) {
            this.source = source == null ? new ESpec().getSource() : source;
            this.providedContentSource = providedContentSource;
            this.logStorage = logStorage;
        }

        @JsonIgnore
        public boolean isValid() {
            switch (source) {
                case PROVIDE_ARCHIVE_FILE:
                case PROVIDE_ARCHIVE_URL:
                case PROVIDE_GIT_REPO_DIR:
                case DIRECT:
                case PROVIDE_DIR: {
                    if (providedContentSource == null)
                        throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be null if espec.source == "+source);
                    return true;
                }
                default:
                    throw new IllegalArgumentException("Config Error: unsupported espec.source == "+source);
            }
        }

        @JsonProperty
        @Nonnull
        public ESpecSource getSource() {
            return source;
        }

        @JsonProperty
        @Nullable
        public String getProvidedContentSource() {
            return providedContentSource;
        }

        @JsonProperty
        @Nullable
        public List<ESpecLogStorageDetails> getLogStorage() {
            return logStorage;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof ESpec)) return false;
            ESpec eSpec = (ESpec) o;
            return source == eSpec.source &&
                    Objects.equals(providedContentSource, eSpec.providedContentSource) &&
                    Objects.equals(logStorage, eSpec.logStorage);
        }

        @Override
        public int hashCode() {
            return Objects.hash(source, providedContentSource, logStorage);
        }
    }



    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Provision {
        //Allocate, Provision, POA    (skipped if requestRspec method is "already_povisioned")
//        private final long provisionExpireTimeMin;
        private final boolean useGeniEndTimeOptionForCreateSliver;
        private final boolean enabled;

        public Provision() {
//            this.provisionExpireTimeMin = 25;
            this.useGeniEndTimeOptionForCreateSliver = false;
            this.enabled = true;
        }
        @JsonCreator
        public Provision(
//                @Nullable @JsonProperty("provisionExpireTimeMin") Long provisionExpireTimeMin,
                @Nullable @JsonProperty("useGeniEndTimeOptionForCreateSliver") Boolean useGeniEndTimeOptionForCreateSliver,
                @Nullable @JsonProperty("enabled") Boolean enabled
        ) {
//            this.provisionExpireTimeMin = provisionExpireTimeMin == null ? new Provision().getProvisionExpireTimeMin() : provisionExpireTimeMin;
            this.useGeniEndTimeOptionForCreateSliver = useGeniEndTimeOptionForCreateSliver == null ? new Provision().getUseGeniEndTimeOptionForCreateSliver() : useGeniEndTimeOptionForCreateSliver;
            this.enabled = enabled == null ? new Provision().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

//        @JsonProperty
//        public long getProvisionExpireTimeMin() {
//            return provisionExpireTimeMin;
//        }

        @JsonProperty
        public boolean getUseGeniEndTimeOptionForCreateSliver() {
            return useGeniEndTimeOptionForCreateSliver;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Provision)) return false;

            Provision provision = (Provision) o;

            if (useGeniEndTimeOptionForCreateSliver != provision.useGeniEndTimeOptionForCreateSliver) return false;
            return enabled == provision.enabled;
        }

        @Override
        public int hashCode() {
            int result = (useGeniEndTimeOptionForCreateSliver ? 1 : 0);
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class WaitForReady {
        private final long maxTimeMin;
        private final long extraWaitTimeSeconds;
        private final boolean enabled;

        public WaitForReady() {
            this.maxTimeMin = 20;
            this.extraWaitTimeSeconds = 0;
            this.enabled = true;
        }
        @JsonCreator
        public WaitForReady(@Nullable @JsonProperty("maxTimeMin") Long maxTimeMin,
                            @Nullable @JsonProperty("extraWaitTimeSeconds") Long extraWaitTimeSeconds,
                            @Nullable @JsonProperty("enabled") Boolean enabled) {
            this.maxTimeMin = maxTimeMin == null ? new WaitForReady().getMaxTimeMin() : maxTimeMin;
            this.extraWaitTimeSeconds = extraWaitTimeSeconds == null ? new WaitForReady().getExtraWaitTimeSeconds() : extraWaitTimeSeconds;
            this.enabled = enabled == null ? new WaitForReady().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxTimeMin < 0)
                throw new IllegalArgumentException("Config Error: waitForReady.maxTimeMin < 0 makes no sense");

            return true;
        }


        @JsonProperty
        public long getMaxTimeMin() {
            return maxTimeMin;
        }


        @JsonProperty
        public long getExtraWaitTimeSeconds() {
            return extraWaitTimeSeconds;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof WaitForReady)) return false;

            WaitForReady that = (WaitForReady) o;

            if (maxTimeMin != that.maxTimeMin) return false;
            if (extraWaitTimeSeconds != that.extraWaitTimeSeconds) return false;
            return enabled == that.enabled;
        }

        @Override
        public int hashCode() {
            int result = (int) (maxTimeMin ^ (maxTimeMin >>> 32));
            result = 31 * result + (int) (extraWaitTimeSeconds ^ (extraWaitTimeSeconds >>> 32));
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof ExperimentSetupConfig)) return false;
        ExperimentSetupConfig that = (ExperimentSetupConfig) o;
        return overrideESpecRSpec == that.overrideESpecRSpec &&
                runLinkTest == that.runLinkTest &&
                Objects.equals(slice, that.slice) &&
                Objects.equals(requestRSpec, that.requestRSpec) &&
                Objects.equals(eSpec, that.eSpec) &&
                Objects.equals(provision, that.provision) &&
                Objects.equals(waitForReady, that.waitForReady);
    }

    @Override
    public int hashCode() {
        return Objects.hash(slice, requestRSpec, eSpec, provision, waitForReady, overrideESpecRSpec, runLinkTest);
    }
}
